/*! file mainform.cpp
*******************************************************************************
File: mainform.cppn
Implementation of the MainForm class.
This file is part of
%URBI Lab, version __urbilabversion__n
(c) Jean-Christophe Baillie, 2004-2005.
Permission to use, copy, modify, and redistribute this software for
non-commercial use is hereby granted.
This software is provided "as is" without warranty of any kind,
either expressed or implied, including but not limited to the
implied warranties of fitness for a particular purpose.
For more information, comments, bug reports: http://urbi.sourceforge.net
**************************************************************************** */
#include <qapplication.h>
#include <qvariant.h>
#include <qpushbutton.h>
#include <qsplitter.h>
#include <qtabwidget.h>
#include <qwidget.h>
#include <qcombobox.h>
#include <qlayout.h>
#include <qtooltip.h>

#include <qmenubar.h>
#include <qimage.h>
#include <qpixmap.h>
#include <qdialog.h>
#include <qlineedit.h>
#include <qstring.h>
#include <qsound.h>
#include <qsettings.h>
#include <qmessagebox.h>
   //Added by qt3to4:
#include <QHBoxLayout>
#include <QMenuItem>
#include <QTimer>
#include "mainform.hh"
#include "umsgtab.hh"
#include "umsgwindow.hh"
#include "udebugwindow.hh"
#include "urobotmng.hh"
#include "urobot.hh"
#include "uhostnameform.hh"

   //! MainForm constructor.
   /*!
     Constructs a MainForm as a child of 'parent', with the
     name 'name' and widget flags set to 'fl'.
     Main window allows you open message and debug connections,
     graph double variables (in the future), play video and sound (in the future)
     from the robot.
     param parent is this windows's parent in Qt hierarchy
     param name within Qt
     param fl are Qt flags
   */
   MainForm::MainForm( QWidget* parent)
     : QMainWindow( parent ) {
       loadSettings();
       robotMng = new URobotMng();
       hostNameForm = 0;
       // Qt widgets initialisation
       (void)statusBar();
       //setOpaqueMoving( TRUE );
       setCentralWidget( new QWidget( this ) );
       MainFormLayout = new QHBoxLayout( centralWidget());

       mainFormSplitter = new QSplitter( centralWidget() );
       mainFormSplitter->setOrientation( Qt::Vertical );

       // client and debug windows
       msgWindow = new UMsgWindow( mainFormSplitter );
       debugWindow = new UDebugWindow( mainFormSplitter );

       MainFormLayout->addWidget( mainFormSplitter );

       // actions
       fileNewMsgAction = new QAction(tr("&New URBI connection..."), this);     
       fileNewDebugAction = new QAction(tr("New connection to &Debug port..."), this);  
       fileCloseMsgAction = new QAction(tr("&Close connection"), this);      
       fileCloseDebugAction = new QAction(tr("Close debug connection"), this);
       fileExitAction = new QAction(tr("E&xit"), this);

       helpAboutAction = new QAction(tr("&About..."), this );

       viewToggleDebugVisibilityAction = new QAction(tr("Show/Hide debug panel"),  this);
       viewToggleDebugVisibilityAction->setCheckable( true );
       viewToggleDebugVisibilityAction->setChecked( false );

       fileNewMsgAction->setShortcut(tr("Ctrl+N"));                                     
       fileNewDebugAction->setShortcut(tr("Ctrl+Shift+N"));    
       fileCloseMsgAction->setShortcut(tr("Ctrl+Q"));
       fileCloseDebugAction->setShortcut(tr("Ctrl+Shift+Q"));
       fileExitAction->setShortcut(tr("Ctrl+X"));
       viewToggleDebugVisibilityAction->setShortcut(tr("Ctrl+D"));                                   
       fileMenu = new QMenu(tr( "&File" ),  this );
       fileMenu->addAction(fileNewMsgAction);
       fileMenu->addAction(fileNewDebugAction);
       fileMenu->addSeparator();
       fileMenu->addAction(fileCloseMsgAction);
       fileMenu->addAction(fileCloseDebugAction);
       fileMenu->addSeparator();
       fileMenu->addAction(fileExitAction);
      
       menuBar()->addMenu( fileMenu );
      
       viewMenu = new QMenu(tr("&View"),  this );
       viewMenu->addAction(viewToggleDebugVisibilityAction);
       menuBar()->addMenu( viewMenu);
       helpMenu = new QMenu(tr("&Help"),  this );

     
       helpMenu->addAction(helpAboutAction);
       menuBar()->addMenu( helpMenu);

       languageChange();

       setGeometry( initX, initY, initWidth, initHeight );
       //clearWState( WState_Polished );

       // signals and slots connections
       connect( fileNewMsgAction, SIGNAL( activated() ), this, SLOT( fileNewMsg() ) );
       connect( fileNewDebugAction, SIGNAL( activated() ), this, SLOT( fileNewDebug() ) );
        connect( fileCloseMsgAction, SIGNAL( activated() ), this, SLOT( fileCloseMsg() ) );
       connect( fileCloseDebugAction, SIGNAL( activated() ), this, SLOT( fileCloseDebug() ) );
       connect( fileExitAction, SIGNAL( activated() ), this, SLOT( fileExit() ) );
  
        connect( helpAboutAction, SIGNAL( activated() ), this, SLOT( helpAbout() ) );
       connect( viewToggleDebugVisibilityAction, SIGNAL( toggled(bool) ), this, SLOT( viewToggleDebugVisibility() ) );
       QTimer::singleShot(100, this, SLOT( fileNewMsg()));
     }


//! MainForm destructor.
/*! Destroys the object and frees any allocated resources. */
MainForm::~MainForm() {
  // no need to delete child widgets, Qt does it all for us

  saveSettings();
}



//! Opens a new client for a robot and adds a tab to msg window.
void
MainForm::fileNewMsg() {
  qDebug("MainForm::fileNewMsg");
  // pop up dialog and ask for robot name
  if( hostNameForm == 0)
    hostNameForm = new UHostNameForm( this );

  // Cancel hit
  if( hostNameForm->exec() == QDialog::Rejected )
    return;

  // TODO: add more extensive control of hostname
  if( hostNameForm->hostNameLineEdit->text().isEmpty() )
    return;

  // ask robot mng to set up a robot with given hostname
  URobot* robot = robotMng->getRobot( hostNameForm->hostNameLineEdit->text() );

  // add tab to msg window
  msgWindow->addNewTab( robot );

  if( ! msgWindow->isVisible() )
    msgWindow->show();
}

//! Opens a new debug for a robot and adds a tab to debug window.
void
MainForm::fileNewDebug() {

  // pop up dialog and ask for robot name
  if( hostNameForm == 0)
    hostNameForm = new UHostNameForm( this );

  // takes robot name from current msg client as a hint
  UMsgTab* currentTab;
  if( ( currentTab = (UMsgTab*)msgWindow->currentWidget() ) != 0 )
    hostNameForm->hostNameLineEdit->setText( currentTab->getRobotHostName() );

  // Cancel hit
  if( hostNameForm->exec() == QDialog::Rejected )
    return;

  if( hostNameForm->hostNameLineEdit->text().isEmpty() )
    return;

  // ask robot mng to set up a robot with given hostname
  URobot* robot = robotMng->getRobot( hostNameForm->hostNameLineEdit->text() );

  // if client exists bring the tab to front, otherwise set up new
  if( robot->debugClientExists() )
    debugWindow->setCurrentPageByPtr( robot->getDebugTab() );
  else
    debugWindow->addNewTab( robot );

  // show debug if hidden
  if( ! debugWindow->isVisible() )
    viewToggleDebugVisibilityAction->setChecked( true );

}


//! Closes current msg client.
void
MainForm::fileCloseMsg() {

  if( ! msgWindow->isVisible() || msgWindow->count() == 0 )
    return;

  //qDebug("MainForm::fileCloseMsg: closing current msg tab...");

  msgWindow->closeCurrentTab( );

  // if it was last tab closed, hide the msg window
  if( msgWindow->isVisible() && msgWindow->count() == 0 )
    msgWindow->hide();
}

//! Closes current debug client.
void
MainForm::fileCloseDebug() {

  if( ! debugWindow->isVisible() || debugWindow->count() == 0 )
    return;

  //qDebug("MainForm::fileCloseDebug: closing current debug tab...");

  debugWindow->closeCurrentTab( );

  // if it was last tab closed, hide the debug window
  if( debugWindow->isVisible() && debugWindow->count() == 0 ) {
    viewToggleDebugVisibilityAction->setChecked( false );
    debugWindow->hide();
  }

}

//! Closes the application.
void
MainForm::fileExit() {
  qApp->closeAllWindows();
}



//! Shows/hides debug window.
void
MainForm::viewToggleDebugVisibility(){

  if( ! debugWindow || debugWindow->count() == 0 )
    return;

  if( debugWindow->isVisible() )
    debugWindow->hide();
  else
    debugWindow->show();
}


void
MainForm::helpAbout() {
  QMessageBox::about(this, "urbilab",
		     "urbilab 0.2\n"
		     "Copyright @2005, ENSTA\n"
		     "http://www.urbiforge.com\n"
		     "Additional feature: type '#graph variablename' to graph the value of a variable over time,\n"
		     "                    or to display a video if this variable is an image\n\n"
		     " Graphing by qwt (http://qwt.sf.net)\n"
		     " This program is distributed under the terms of the GNU GPL license"
		     );
}
//! Loads user settings.
void
MainForm::loadSettings() {

  settings.beginGroup( APP_KEY + "/geometry" );
  initX      = settings.value( "/x",      50 ).toInt();
  initY      = settings.value( "/y",      50 ).toInt();
  initWidth  = settings.value( "/width",  640 ).toInt();
  initHeight = settings.value( "/height", 480 ).toInt();
  settings.endGroup();
}
//! Saves user settings.
void
MainForm::saveSettings() {
  settings.beginGroup( APP_KEY + "/geometry" );
  settings.setValue( "/x" ,     x() );
  settings.setValue( "/y" ,     y() );
  settings.setValue( "/width" , width() );
  settings.setValue( "/height", height() );

  settings.endGroup();


}
