/*! file udebugwindow.cpp
*******************************************************************************

File: udebugwindow.cppn
Implementation of the UDebugWindow class.

This file is part of
%URBI Lab, version __urbilabversion__n
(c) Jean-Christophe Baillie, 2004-2005.

Permission to use, copy, modify, and redistribute this software for
non-commercial use is hereby granted.

This software is provided "as is" without warranty of any kind,
either expressed or implied, including but not limited to the
implied warranties of fitness for a particular purpose.

For more information, comments, bug reports: http://urbi.sourceforge.net

**************************************************************************** */

#include <string.h>

#include <qstring.h>
#include <qwidget.h>

#include "udebugwindow.hh"
#include "udebugtab.hh"
#include "urobot.hh"
#include "umsgtab.hh"

   //! UDebugWindow constructor.
   /*!
     Construct the widget that will hold debug tabs and hides it.

     param parent is this windows's parent in Qt hierarchy
     param name within Qt

   */
   UDebugWindow::UDebugWindow( QWidget* parent)
     : QTabWidget( parent) {

       // there are no debug tabs yet, so we hide the window
       if( count() == 0 )
	 hide();
     }


//! UDebugWindow destructor.
UDebugWindow::~UDebugWindow() {
  // no need to delete anything, Qt does it for us
}


//! Brings to front debug tab for particular robot.
/*!
  param hostName is the hostname of the robot whose debug tab should be brought to the front

*/
void
UDebugWindow::toFront( const char* hostName ) {
  setCurrentIndex( pageIndexByHostName( hostName ) );
}


//! Brings to front particular tab.
/*!
  param page is the pointer to the tab we want to bring to the front
*/
void
UDebugWindow::setCurrentPageByPtr( UDebugTab *page ) {
  setCurrentIndex( indexOf( (QWidget*)page ) );
}


//! Adds a new tab to the window.
/*!
  param robot is the URobot object that the tab will be using
  when communicating with the robot

*/
void
UDebugWindow::addNewTab( URobot *robot ) {
  // creates new tab
  UDebugTab* tab = new UDebugTab( this );

  // interconnects robot and tab
  tab->setRobot( robot );
  robot->setupDebugClient( tab );

  // adds the tab to the window
  addTab( tab, QString(robot->getHostName())+":debug" );

  // brings the tab to the front
  setCurrentIndex( indexOf( tab ));
}


//! Removes current tab from the window and destroys it.
void
UDebugWindow::closeCurrentTab( ) {
  if( count() == 0 )
    return;

  QWidget* tab = currentWidget();
  int index = currentIndex();
  qDebug( "UDebugWindow::closeCurrentTab: removing and deleting current tab..." );

  removeTab( index );
  delete tab;

}


//! Returns tab's index according to robot's name
/*!
  param hostName is the hostname of the robot whose tab's index we ask
*/
int
UDebugWindow::pageIndexByHostName( const char* hostName ) {
  QWidget* tab;
  for( int i = 0; i < count(); i++ ) {
    tab = widget( i );
    if( strcmp( ((UMsgTab *)tab)->getRobotHostName(), hostName ) == 0 )
      return i;
  }
  return -1;
}



//! Returns UDebugTab object according to robot's name
/*!
  param hostName is the hostname of the robot whose tab's pointer we ask
*/
QWidget*
UDebugWindow::pageByHostName( const char* hostName ) {
  return widget( pageIndexByHostName( hostName ) );
}



