/*
 *   ugraph.cpp
 *
 *   Implementation of UGraph
 *   desc: class that provides graphing and stores all current graphing processes
 *         by graphing we mean presenting to the user double or video or sound variables
 */

#include <qapplication.h>
#include <qevent.h>

#include "ugraph.hh"
#include "ugraphwindow.hh"
#include "udoublegraphwindow.hh"
#include "urobot.hh"
#include "ucustomevent.hh"


UGraph::UGraph()
  : QObject() {

  }

UGraph::~UGraph() {
}

bool
UGraph::event( QEvent* e) {
  UCustomEvent* ce = dynamic_cast<UCustomEvent*>( e );

  switch( e->type() ) {

  case UCustomEvent::EVENT_PLOT_DOUBLE:
    plotDouble( *(ce->msg) );
    break;
  case UCustomEvent::EVENT_PLOT_IMAGE:
    plotImage( *(ce->msg) );
    break;
  case UCustomEvent::EVENT_PLOT_SOUND:
    plotSound( *(ce->msg) );
    break;
  default:
    break;

  }

  return QObject::event( e );
}

void
UGraph::registerGraphProcess( URobot* robot, QString& uniqueTag, QString& var,
			      UMessageType msgType, UBinaryMessageType binType ) {
  
 
  qDebug( "UGraph::registerGraphProcess: registering  %s (%d,%d) with tag %s" , var.toAscii().constData(), msgType, binType, uniqueTag.toAscii().constData() );

  graphProcessList.insert( uniqueTag, new UGraphProcess( robot, var, msgType, binType ) );

  //QDictIterator<UGraphProcess> it( graphProcessList );
  //GraphProcess* process;

  // while( ( process = it.current() ) != 0 ) {
  //     qDebug("uniqueTag: " );
  //     qDebug(it.currentKey());
  //     ++it;
  //   }
}

UCallbackAction
UGraph::plotEventDouble( const UMessage& msg ) {
  UCustomEvent* e = new UCustomEvent( (QEvent::Type) UCustomEvent::EVENT_PLOT_DOUBLE );
  e->msg = new UMessage( msg ); //deep copy

  qApp->postEvent( this, e );

  return URBI_CONTINUE;
}

UCallbackAction
UGraph::plotEventImage( const UMessage& msg ) {
  UCustomEvent* e = new UCustomEvent( (QEvent::Type) UCustomEvent::EVENT_PLOT_IMAGE );
  e->msg = new UMessage( msg ); //deep copy

  qApp->postEvent( this, e );

  return URBI_CONTINUE;
}

UCallbackAction
UGraph::plotEventSound( const UMessage& msg ) {
  UCustomEvent* e = new UCustomEvent( (QEvent::Type) UCustomEvent::EVENT_PLOT_SOUND );
  e->msg = new UMessage( msg ); //deep copy

  qApp->postEvent( this, e );

  return URBI_CONTINUE;
}

void
UGraph::plotDouble( UMessage& msg ) {
  if( msg.type != MESSAGE_DOUBLE ) {
    qDebug("UGraph::plotDouble: !!! type of variable changed...graphing stopped..." );
    stopGraph( QString( msg.tag ) );
    //TODO: stop graphing and warn user that type of variable changed
    return;
  }

  //qDebug( "UGraph::plotDouble: received %s:%f ", msg.tag, msg.doubleValue );
  UGraphProcess *process = graphProcessList[msg.tag];

  if( !process ) {
    qWarning( "UGraph::plotImage: graphing process with tag %s is not registered with graph object", msg.tag );
    return;
  }

  if( msg.type != MESSAGE_DOUBLE ) {
    qDebug("UGraph::plotImage: !!! type of variable changed...graphing stopped..." );
    stopGraph( QString( msg.tag ) );
    //TODO: warn user that type of variable changed
    return;
  }
#ifndef NO_DOUBLEPLOT
  UGraphWindow *window = (UDoubleGraphWindow* )process->window;

  if( window == NULL ) {  // first time: we construct the window
    process->window = window = new UDoubleGraphWindow();
    process->window->initialize(process->robot, process->var, msg.tag);
    process->window->show();
  }

  if( window->isVisible( ) ) { // if window is visible we update, otherwise it is not necessary
    if( !window->updateGraph( msg ) ) { // if update hasnt succeeded we stop the graphing and tell the user
      //TODO: tell user
      stopGraph( QString( msg.tag ) );
    }
    process->window->repaint();
  }
  
  #endif
}

void
UGraph::plotImage( UMessage& msg ) {
  UGraphProcess *process = graphProcessList[msg.tag];

  if( !process ) {
    qWarning( "UGraph::plotImage: graphing process with tag %s is not registered with graph object", msg.tag );
    return;
  }

  if( msg.type != MESSAGE_BINARY || msg.binaryType != BINARYMESSAGE_IMAGE ) {
    qDebug("UGraph::plotImage: !!! type of variable changed...graphing stopped..." );
    stopGraph( QString( msg.tag ) );
    //TODO: warn user that type of variable changed
    return;
  }

  UGraphWindow *window = process->window;

  if( window == NULL ) {  // first time: we construct the window
    process->window = window = new UGraphWindow();
    process->window->initialize(process->robot, process->var, msg.tag);
    process->window->show();
  }

  if( window->isVisible( ) ) { // if window is visible we update, otherwise it is not necessary
    if( !window->updateGraph( msg ) ) { // if update hasnt succeeded we stop the graphing and tell the user
      //TODO: tell user
      stopGraph( QString( msg.tag ) );
    }
    process->window->repaint();
  }

}

void
UGraph::plotSound( UMessage& msg ) {
  if( msg.type != MESSAGE_BINARY || msg.binaryType != BINARYMESSAGE_SOUND ) {
    qDebug("UGraph::plotSound: !!! type of variable changed...graphing stopped..." );
    stopGraph( QString( msg.tag ) ); //TODO
    //TODO: stop graphing and warn user that type of variable changed
    return; // TODO : removing callback already removed by stopGraph()
  }

  qDebug( "UGraph::plot:Sound received %s:%d ", msg.tag, msg.sound.size );

}

void
UGraph::stopGraph( QString uniqueTag ) {
  qDebug( "UGraph::stopGraph: stopping graphing associated wit tag %s", uniqueTag.toAscii().constData() );
  UGraphProcess* process = graphProcessList[uniqueTag];

  if( !process ) {
    qWarning( "UGraph::stopGraph: trying to stop graphing that is not registered..." );
    return;
  }

  // we tell the robot to stop
  process->robot->stopGraph( uniqueTag );

  // we remove the process from the list
  delete graphProcessList.take( uniqueTag );


}

void
UGraph::stopGraph( UGraphWindow* window ) {
  for ( QHash<QString, UGraphProcess*>::iterator it = graphProcessList.begin();
        it != graphProcessList.end(); it++) 
           if ((*it)->window == window){
              stopGraph(it.key());
              return;
              }
  
}

