/*! file umsgoutpane.cpp
*******************************************************************************

File: umsgoutpane.cppn
Implementation of the UMsgOutPane class.

This file is part of
%URBI Lab, version __urbilabversion__n
(c) Jean-Christophe Baillie, 2004-2005.

Permission to use, copy, modify, and redistribute this software for
non-commercial use is hereby granted.

This software is provided "as is" without warranty of any kind,
either expressed or implied, including but not limited to the
implied warranties of fitness for a particular purpose.

For more information, comments, bug reports: http://urbi.sourceforge.net

**************************************************************************** */
#include <QTextEdit>
#include <QDir>
#include <QFile>
#include <qstring.h>
#include <qsound.h>
#include <qlineedit.h>

#include "mainform.hh"
#include "umsgoutpane.hh"
#include "ucmdline.hh"
#include "umsgtab.hh"
#include "ugraph.hh" //TODO: remove, graph command will be submitted from outside
#include "urotextedit.hh"

   //! UMsgOutPane constructor.
   /*! Sets up a text field and a command line.

   param parent is this widget's parent in Qt hierarchy
   param name within Qt
   */
   UMsgOutPane::UMsgOutPane( QWidget* parent)
     : QWidget( parent) {
       

       QVBoxLayout *layout = new QVBoxLayout;
       cmdLine = new UCmdLine( this);
       msgOutTextEdit = new UROTextEdit( this, cmdLine);
       msgOutTextEdit->setFontFamily("Courier");
       msgOutTextEdit->setFontPointSize(10);
       msgOutTextEdit->setFrameShadow( QTextEdit::Sunken );
       msgOutTextEdit->setReadOnly( TRUE );
       msgOutTextEdit->setUndoRedoEnabled( FALSE );
       

       layout->addWidget(msgOutTextEdit);
       layout->addWidget(cmdLine);
       setLayout(layout);
       // loads user settings
       loadSettings();

       // when Enter pressed send() is called
       connect( cmdLine->lineEdit(), SIGNAL(returnPressed()), this, SLOT(send()) );

     }

//! UMsgOutPane destructor.
UMsgOutPane::~UMsgOutPane() {
  // no need to delete anything, Qt does everything

  // saves user settings
  saveSettings();
}

//! Sets the msgTab pointer to the value passed in.
/*!
  param tab
*/
void
UMsgOutPane::setMsgTab( UMsgTab* tab ) {
  msgTab = tab;
}

//! Forwards send call to msgTab.
/*! If not empty, it forwards usrCmd to message tab.
  param usrCmd
*/
void
UMsgOutPane::send( const QString &usrCmd ) {

  // empty command is not sent
  if( usrCmd.isEmpty() ) {
    qDebug("UMsgOutPane::send: Trying to send an empty command...");
    return;
  }

  QString command = usrCmd;
  
  // puts the command in the message history
  msgOutTextEdit->append( command );
  if (cmdLine->count() >1)
    cmdLine->removeItem(cmdLine->count()-2);
  //put the message in double in combo box 
  //since up-arrow in combobox sends you to n-1 th message.
  cmdLine->setCurrentIndex(cmdLine->count()-1);
  cmdLine->addItem( command );
  cmdLine->setCurrentIndex(cmdLine->count()-1);

  /************** TODO:remove ************/
  // pseudocode parsing
  if( command == "#test sound;" || command == "#test sound") {
    qDebug("sound device %s available", QSound::isAvailable() ? "is" : "ain't");
    QSound::play("/usr/share/tuxracer/sounds/tux_on_ice1.wav");
    return;
  }
  else if( command.startsWith( "#graph " ) ) {

    QString variable = command.mid( 7, command.length() - 7  ); // we cut graph "#graph "
    while (variable.contains( ";" ) || variable.contains( "," ))
      variable = variable.mid(0, variable.length()-1);
    // if it is not a global variable ("." - notation) we prepend connection id
    if( !variable.contains( "." ) )
      variable.prepend( "." ).prepend( msgTab->getConnId( ) ); // creates ConnId.variable_name

    qDebug("UMsgOutPane::send: variable to graph %s" , variable.toAscii().constData());

    msgTab->getRobot( )->startGraph( variable ); // graph this variable //TODO: remove

  }
  else if( command.startsWith( "#stopgraph " ) ) {
    QString tag = command.mid( 11, command.length() - 11); // we cut graph "#stopgraph "

    qDebug("UMsgOutPane::send: request for stopping graphing assoc. with tag %s" ,tag.toAscii().constData());

    graph.stopGraph( tag ); //TODO: remove
  }
  /****************** end TODO:remove**************************/
  else {
    // sends the command to the parent msgTab
    msgTab->send( command );
  }

  // clears the command line
  cmdLine->lineEdit()->clear();
}

//! Convinience function, similar as above.
/*! Reads the command from command line and calls
  UMsgOutPane::send( const QString &usrCmd ).
*/
void
UMsgOutPane::send( ) {
  send( cmdLine->lineEdit()->text() );
}

//! Loads user settings
void
UMsgOutPane::loadSettings( ) {
  //load history
  startIndex=0;
  QString history = QDir::homePath()+QString("/.urbilab_history");
  QFile f(history);
  if (!f.exists())
    return;
  f.open(QIODevice::ReadOnly);
  char *line = new char[2048];
  while (f.readLine(line, 2048)!=-1) {
    if (line[0] && line[strlen(line)-1]=='\n')
      line[strlen(line)-1]=0;
    cmdLine->addItem(line);
  }
  cmdLine->addItem(""); 
  cmdLine->setCurrentIndex(cmdLine->count()-1);
  startIndex = cmdLine->count()-1;
  delete []line;
}

//! Saves user settings.
void
UMsgOutPane::saveSettings( ) {


  //save history
  QString history = QDir::homePath()+"/.urbilab_history";
  QFile f(history);
  f.open(QIODevice::Append);
  for (int i=startIndex; i<cmdLine->count()-1;i++) {
    f.write(cmdLine->itemText(i).toAscii());
    f.write("\n");
  }
  f.close();
}
