/*! file umsgtab.cpp
*******************************************************************************

File: umsgtab.cppn
Implementation of the UMsgTab class.

This file is part of 
%URBI Lab, version __urbilabversion__n
(c) Jean-Christophe Baillie, 2004-2005.

Permission to use, copy, modify, and redistribute this software for
non-commercial use is hereby granted.

This software is provided "as is" without warranty of any kind,
either expressed or implied, including but not limited to the
implied warranties of fitness for a particular purpose.

For more information, comments, bug reports: http://urbi.sourceforge.net

**************************************************************************** */

#include <qlayout.h>
#include <qsplitter.h>

#include <QTextBrowser>
#include <QScrollBar>
#include <QUrl>
#include <QFileDialog>
#include <QMessageBox>
#include <QTimer>
#include <qstring.h>
#include <qmime.h>
#include <qimage.h>
#include <qsound.h>
#include <qfile.h>
#include <qmap.h>
#include <qsettings.h>
   //Added by qt3to4:
#include <QVBoxLayout>
#include <QEvent>
#include <QCoreApplication>
#include <uclient.h> //liburbi

#include "umsgtab.hh"
#include "ufunctions.hh"
#include "umsginpane.hh"
#include "umsgoutpane.hh"
#include "ucustomevent.hh"


   UCacheTextBrowser::UCacheTextBrowser(QWidget * parent, QWidget * dest) 
     : QTextBrowser(parent), dest(dest){
       resources = new QTextDocument();
     }



void UCacheTextBrowser::keyPressEvent ( QKeyEvent * e ) {
  if (e->modifiers() & Qt::ControlModifier)
    QTextEdit::keyPressEvent(e);
  else {
    dest->setFocus(Qt::OtherFocusReason);
    QCoreApplication::sendEvent(dest,e);
  }
}

QVariant UCacheTextBrowser::loadResource ( int type, const QUrl & name ) {
  return resources->resource(type, name);
}

QTextDocument * UCacheTextBrowser::cache() {
  return resources;
}


   //! UMsgTab constructor.
   /*! Sets up widgets containing the both textfields and command line.

   param parent is this widget's parent in Qt hierarchy
   param name within Qt
   */
   UMsgTab::UMsgTab( QWidget* parent)
     : QWidget( parent ) {

       msgTabLayout = new QVBoxLayout( this);

       msgPaneSplitter = new QSplitter( this);
       msgPaneSplitter->setOrientation( Qt::Vertical );

       msgTabLayout->addWidget(msgPaneSplitter);

       msgOutPane = new UMsgOutPane( msgPaneSplitter);
       msgInPane = new UMsgInPane( msgPaneSplitter, (QWidget *)msgOutPane->getCmdLine());
       msgPaneSplitter->addWidget(msgInPane);
       msgPaneSplitter->addWidget(msgOutPane);
       msgOutPane->setMsgTab( this );
       msgInTextEdit = msgInPane->getTextEdit();

       // connect clicking with a function that checks if a line with an image was clicked
       //connect( msgInTextEdit, SIGNAL(clicked(int,int)), this, SLOT(textEditClicked(int,int)));
       connect( msgInTextEdit, SIGNAL(anchorClicked(const QUrl)), this, SLOT(anchorClicked(const QUrl)));

       // initialise image and sound name suffices
       imgId = 1;  
       soundId = 1;

       // set up the tmp dir path
       tmpDirName = getTempDir();
       startTime = QTime::currentTime();
     }


//! UMsgTab destructor.
/*! Destroys the tab and makes the respective URobot object close
  the client used by this tab.

  Saves settings.
*/
UMsgTab::~UMsgTab() {

  qDebug("UMsgTab::~UMsgTab: destroying tab...");

  // saves settings
  saveSettings();

  // closes the corresponding client for this tab
  robot->closeMsgClient( this ); 


}


//! Handles events sent to this object.
/*! If a custom event of type EVENT_RECEIVE_MSG is received
  it forwards the UMessage to the receiveMsg function.

  param e is the QEvent object. We use UCustomEvent defined in ucustomevent.hh
*/ 
bool 
UMsgTab::event( QEvent* e ) {

  switch( e->type() ){
  case UCustomEvent::EVENT_RECEIVE_MSG: {
    UCustomEvent* ce = dynamic_cast<UCustomEvent*>( e );
    receiveMsg( *( ce->msg ) );
    break;
  }
  default:
    break;
  }

  // handling is passed along, deletion of QEvent object inside
  return QWidget::event( e );
}


//! Sends the command to the robot.
/*! For the robot to distiguish between several tabs it sends this pointer.

param command is the URBI command sent to the robot
*/ 
void 
UMsgTab::send( QString& command ) {
  robot->send( this, command );
}



//! Receives, parses and displays, resp. plays message.
/*! 
  According to message type does following actions:
  displays all messages in msgInTextEdit except for that of type MESSAGE_BINARY
  message of type MESSAGE_BINARY converts to appropriate sound or image format
  and then plays or displayes

  For binary messages of type BINARYMESSAGE_IMAGE and BINARYMESSAGE_UNKNOWN
  offers the saving possibility as an anchor in the text field

  param msg to receive
*/
void 
UMsgTab::receiveMsg( UMessage& msg ) {
#ifndef QT_NO_DEBUG
  QMap<int, QString> msgType; // list of message type names, debug purposes

  msgType.insert( MESSAGE_DOUBLE,  "DOUBLE" );
  msgType.insert( MESSAGE_STRING,  "STRING" );
  msgType.insert( MESSAGE_BINARY,  "BINARY" );
  msgType.insert( MESSAGE_SYSTEM,  "SYSTEM" );
  msgType.insert( MESSAGE_ERROR,   "ERROR" );
  msgType.insert( MESSAGE_UNKNOWN, "UNKNOWN" );

  qDebug("UMsgTab::receiveMsg: received msg of type %s", msgType[msg.type].toAscii().constData() );
#endif // #ifdef QT_DEBUG

  QString prefix; // message time and tag
  prefix += (QString)"[" + QString::number( msg.timestamp ) + (QString)":" + (QString)msg.tag + (QString)"] ";

  /*
  QScrollBar *sb =  msgInTextEdit->verticalScrollBar();
  bool sbDown = sb->value() == sb->maximum();
  */
  
  /**** parsing ***/
  switch(msg.type) {
    /***** double ****/
  case MESSAGE_DOUBLE: 
    msgInTextEdit->append( prefix + QString::number( msg.doubleValue, 'f', 6 ) );
    break;

    /***** string ****/
  case MESSAGE_STRING:
    msgInTextEdit->append( prefix + (QString)msg.stringValue );
    break;

    /***** binary ****/
  case MESSAGE_BINARY: {

    switch(msg.binaryType) {
      /***** bin none ****/ 
    case BINARYMESSAGE_NONE:
      //will never happen
      qWarning("UMsgTab::receiveMsg: !!! MESSAGE_BINARY type is BINARYMESSAGE_NONE");
      break;

      /***** bin unknown ****/
    case BINARYMESSAGE_UNKNOWN:
      msgInTextEdit->append( prefix 
			     + "BIN " 
			     + QString::number(msg.binary.size) );
      //TODO
      // offer to save data
      break; 

      /***** bin image ****/
    case BINARYMESSAGE_IMAGE: {
      uchar* buffer32 = 0;                // buffer for 32-bit RGB

      QMap<int, QString> formatName;      // names of image formats, debug purposes
      formatName.insert( IMAGE_JPEG,  "jpeg" );
      formatName.insert( IMAGE_YCbCr, "YCbCr" );
      formatName.insert( IMAGE_RGB,   "rgb" );
      formatName.insert( IMAGE_PPM,   "ppm" );

      // img name is img_robotname-connordernumber_imgordernumber
      QString imgName("img_"               
		      + QString( getRobotHostName() ).replace(":", "-") + "_"
		      + QString::number( imgId++ ) ); 

      // displaying image header
      msgInTextEdit->append( prefix 
			     + "BIN " 
			     + QString::number(msg.image.size) + " "
			     + formatName[msg.image.imageFormat] + " "
			     + QString::number(msg.image.width) + " "
			     + QString::number(msg.image.height) );


      uchar* buffer24 = new uchar[msg.image.width * msg.image.height*3];
      int bsz = msg.image.width * msg.image.height*3;
      if( msg.image.imageFormat == IMAGE_JPEG )
	convertJPEGtoRGB( (const byte*)msg.image.data, msg.image.size, (byte*)buffer24, bsz);
      else if ( msg.image.imageFormat == IMAGE_YCbCr)
	convertYCrCbtoRGB( (const byte*)msg.image.data, msg.image.size, (byte*)buffer24 );

      // converts 24-bit BGR to 32-bit RGB
      buffer32 = new uchar[bsz + bsz / 3]();

      for( int i = 0; i < bsz / 3; i++){
	buffer32[4 * i] = buffer24[3 * i + 2];     // swap R and B
	buffer32[4 * i + 1] = buffer24[3 * i + 1];
	buffer32[4 * i + 2] = buffer24[3 * i + 0]; // swap R and B
	buffer32[4 * i + 3] = 0xff;                // pad by 0s
      }

      // constructs image and fills with data 
      QImage image( buffer32, msg.image.width, msg.image.height, QImage::Format_RGB32);

      msgInTextEdit->cache()->addResource(QVariant::Image, imgName, image);
      
      delete [] buffer24;


      // assigning an image number to a line number, used by save-image workaround
      //lineNum2ImgName.insert( msgInTextEdit->lines(), imgName );

      // adds img tag to the text field
      QString content =  "<img source=\"" + imgName + "\"><a href=\"" + imgName + "\">Save As..</a>";
      msgInTextEdit->append(content );

      //qDebug("UMsgTab::receiveMsg: image %s inserted to msgInTextEdit on line %d", imgName.ascii(), msgInTextEdit->lines() - 1 );

    } // case BINARYMESSAGE_IMAGE:
      break; 


      /***** bin sound ****/
    case BINARYMESSAGE_SOUND: {
      QMap<int, QString> formatName;      // names of sound formats, debug purposes
      formatName.insert( SOUND_RAW, "raw" );
      formatName.insert( SOUND_WAV, "wav" );
      formatName.insert( SOUND_MP3, "mp3" );
      formatName.insert( SOUND_OGG, "ogg" );

      bool sndAvail = true; //QSound::available(); // true if sound device is available

      if( ! sndAvail ) {
	//TODO: popup [u wont be able to play the sound]
	qDebug("UMsgTab::receiveMsg: sound device is not available...");
      }

      // save sound to a temp file, Qt can play sound only from a file
      QFile soundFile( tmpDirName 
		       + "snd_" 
		       + QString( getRobotHostName() ).replace(":", "-")
		       + "_"
		       + QString::number( soundId++ ) + ".wav" );
      qDebug( "UMsgTab::receiveMsg: sound %s received...", (const char *)soundFile.fileName().constData() );


      // displaying image header
      msgInTextEdit->append( prefix 
			     + "BIN " 
			     + QString::number(msg.sound.size) + " "
			     + formatName[msg.sound.soundFormat] + " "
			     + QString::number(msg.sound.channels) + " "
			     + QString::number(msg.sound.rate)  + " "
			     + QString::number(msg.sound.sampleSize) + " "
			     + QString::number(msg.sound.sampleFormat) );

      /*** converting the sound data ***/
      USound tmpSnd;
      tmpSnd.soundFormat = SOUND_WAV;
      tmpSnd.rate = 16000;
      tmpSnd.sampleSize = 16;
      tmpSnd.sampleFormat = SAMPLE_SIGNED;
      tmpSnd.channels = 0;
      tmpSnd.data = 0;
      tmpSnd.size = 0;
      convert( msg.sound, tmpSnd ); // provided by liburbi

      /*** creating file and writing  sound data ***/
      if( ! soundFile.open( QIODevice::WriteOnly ) ) {
	qWarning("UMsgTab::receiveMsg: can't open temp sound file...");
	break;
      }
      if( soundFile.write( tmpSnd.data, tmpSnd.size ) != tmpSnd.size ) {
	qWarning("UMsgTab::receiveMsg: can't write temp sound file...");
	break;
      }
      soundFile.close();

      /*** playing sound and deleting sound file ***/
      if( sndAvail && playSoundOnReceive ) {
	qDebug("UMsgTab::receiveMsg: trying to play %s" , (const char *)soundFile.fileName().constData());
	QSound::play( soundFile.fileName() );
      }

      qDebug("UMsgTab::receiveMsg: deleting %s", (const char *)soundFile.fileName().constData());
      soundFile.remove();

    }
      break;

    default:
      Q_ASSERT(true);
    } // switch(msg.binaryType) {

  } //case MESSAGE_BINARY: {  
    break;

    /***** system ****/
  case MESSAGE_SYSTEM:
    msgInTextEdit->append( prefix + (QString)"***" + (QString)msg.systemValue );
    break;

    /***** error ****/
  case MESSAGE_ERROR:
    msgInTextEdit->append( prefix + (QString)"!!!" + (QString)msg.errorValue );
    break; 

    /***** unknown ****/
  case MESSAGE_UNKNOWN:
    msgInTextEdit->append( prefix + (QString)"???" + (QString)msg.message );
    break;

  default:
    Q_ASSERT(true);

  } // switch(msg.type){
  
  if (startTime.msecsTo(QTime::currentTime()) < 1000) {
    QScrollBar *sb =  msgInTextEdit->verticalScrollBar();
    sb->setValue(sb->maximum());
  }
  /*
  if (sbDown) {
    sb->setValue(sb->maximum());
  }
  */
}



//! Sets pointer to the associated URobot object.
/*!
  param robot is a pointer to URobot object
*/
void        
UMsgTab::setRobot( URobot* robot ) { 
  this->robot = robot; 
};



//! Returns pointer to the associated URobot object.
URobot*     
UMsgTab::getRobot() const { 
  return robot; 
};



//! Returns hostname of the associated URobot object.
const char* 
UMsgTab::getRobotHostName() const { 
  return robot->getHostName(); 
};



//! Returns ID of the associated UClient connection. (Managed by URobot)
char*      
UMsgTab::getConnId() { 
  return robot->getMsgConnIdByMsgTab( this ); 
};



//! Signalizes connection error while URobot was trying to connect to the robot. 
void 
UMsgTab::clientError( URobot::UClientErrType type ) {

  switch( type ) {
  case URobot::MsgClientErr:
    //TODO: popup
    qDebug("UMsgTab::clientError: error opening msg client for tab %s", getRobotHostName());
    break;

  case URobot::GraphClientErr:
    //TODO:popup
    qDebug("UMsgTab::clientError: error opening graph client for tab %s", getRobotHostName());
    break;

  default:
    Q_ASSERT( true );

  }

}


//! Intermediary slot for saving images. Activated by clicked() signal from the text field.
/*! Called when text field with incoming messages is clicked.

Function looks if there is an image on the line that was clicked.
If so it calls saveImage() function with the image name.

param para is line number
param pos is column number
*/
// void
// UMsgTab::textEditClicked( int para, int pos ) {

//   // check if there's an image on the line that was clicked
//   qDebug("UMsgTab::myClicked: clicked %d.paragraph at position %d", para, pos );

//   // if there isn't any image or the click occured on a position behind the image
//   // and the "Save As.." command (pos>9)then return with no action
//   if( lineNum2ImgName.find( para ) == lineNum2ImgName.end() || pos > 9 )
//     return;

//   // otherwise get the image name...
//   QString imgName = lineNum2ImgName[para];

//   qDebug( "UMsgTab::myClicked: saving image " + imgName + " on line %d", para );

//   //....and offer save action to the user
//   saveImage( imgName );

// }


void 
UMsgTab::restoreInPane() {
  qDebug("UMsgTab: restoring inPane to saved state ");
  msgInTextEdit->setHtml(inPaneHTML);
  if (inPanePos !=0)
    msgInTextEdit->verticalScrollBar()->setValue(inPanePos);
}


//! Slot called when an anchor is clicked
/*! Forwards the image name to be saved to fubction saveImg().

Function looks if there is an image on the line that was clicked.
If so it calls saveImage() function with the image name.

param link is name of the image
*/
void 
UMsgTab::anchorClicked( const QUrl &link) {
  QString tmp = link.path();


  qDebug("UMsgTab: saving inPane state ");
  inPaneHTML = msgInTextEdit->toHtml();
  //saving also scrolbar position
  QScrollBar * vs = msgInTextEdit->verticalScrollBar();
  if (!vs)
    inPanePos = 0;
  else
    inPanePos = vs->value(); 
  saveImage( tmp );
}

//! Lets the user save an image according to its name
/*!
  Pops up a file dialog where the user can choose filename and directory
  where the image will be saved.
  He can as well choose from supported image formats.

  param imgName is internal image name

*/
void
UMsgTab::saveImage( QString &imgName ) {
  Q_ASSERT( !imgName.isEmpty() );
  if (imgName[0]=='#')
    imgName.remove(0,1);
  qDebug("UMsgTab::saveImage: saving image %s" , imgName.toAscii().constData());

  // popups up file dialog

  QFileDialog *fd = new QFileDialog(this);
  fd->setModal(true);
  //fd->setTitle("Choose a filename to save under");
  QStringList types;
  types << "PNG files (*.png)" <<
    "JPEG files (*.jpg *.jpeg)" <<
    "BMP files (*.bmp)";
  fd->setFilters(types);
  fd->setConfirmOverwrite(true);
  fd->setAcceptMode(QFileDialog::AcceptSave);
  fd->setFileMode(QFileDialog::AnyFile);
  fd->setDirectory(QDir::current());
  fd->selectFile(imgName + ".png");
  fd->exec();
  QTimer::singleShot(200, this, SLOT(restoreInPane()));
  if(fd->result()==QDialog::Rejected  ) {
    //qDebug("UMsgTab::saveImage: saving of image " + imgName + " canceled..." );
    return;
    }
  QStringList sl = fd->selectedFiles();
  QString imgFileName = sl.front();

  QImage image =  msgInTextEdit->cache()->resource(QVariant::Image, imgName).value<QImage>();
  QString format = fd->selectedFilter();
  format = format.section(' ', 0, 0);
  if (!imgFileName.contains("."))
      imgFileName.append("."+format.toLower());

  qDebug("UMsgTab::saveImage: saving with format %s", format.toAscii().constData());
  delete fd;
  if (!image.save(imgFileName,format.toAscii().constData())) {
    QMessageBox::warning(this, "error saving file", "error saving "+imgFileName+" (unsupported format?)",QMessageBox::Cancel, QMessageBox::NoButton);
    
  }
  else {
    qDebug("UMsgTab::saveImage: image file " + imgFileName.toAscii()+ " saved..." );
  }
 
}

//! Loads user settings
void 
UMsgTab::loadSettings() {
  playSoundOnReceive = settings.value( APP_KEY + "/sound/playOnReceive", true ).toBool();
}

//! Saves user settings.
void 
UMsgTab::saveSettings() {
  settings.setValue( APP_KEY + "/sound/playOnReceive", playSoundOnReceive );
}
