/*! \file  umsgtab.hh
 *******************************************************************************

 File: umsgtab.hh\n
 Definition of the UMsgTab class.

 This file is part of 
 %URBI Lab, version __urbilabversion__\n
 (c) Jean-Christophe Baillie, 2004-2005.

 Permission to use, copy, modify, and redistribute this software for
 non-commercial use is hereby granted.

 This software is provided "as is" without warranty of any kind,
 either expressed or implied, including but not limited to the
 implied warranties of fitness for a particular purpose.

 For more information, comments, bug reports: http://urbi.sourceforge.net

 **************************************************************************** */
/*
 *   umsgtab.hh 
 *
 *   Implementation of UMsgTab
 *   desc: implements tab that includes command line, and text fields for incoming and outgoing messages
 *         it represents one single connection
 */

#ifndef UMSGTAB_H
#define UMSGTAB_H

#include <qwidget.h>
#include <qmap.h>
#include <QTime>
//Added by qt3to4:
#include <QVBoxLayout>
#include <QTextBrowser>
#include <QEvent>

#include "urobot.hh"

class QVBoxLayout;
class QSplitter;
class QString;
class QUrl;
class QTextBrowser;
class QTextDocument;
//class Q3TextEdit;
//class QTextEdit;
class QSettings;
class QString;
class QEvent;

class URobot;
class UClient;
class UMessage;

class UMsgInPane;
class UMsgOutPane;
class UCustomEvent;

extern const QString   APP_KEY;  ///< application key
extern       QSettings settings; ///< contains the user settings


//! Implementation of a QTextBrowser with caching persistant to page changes.
class UCacheTextBrowser: public QTextBrowser {
  
 public:
  UCacheTextBrowser(QWidget * parent, QWidget * dest);
  virtual QVariant loadResource ( int type, const QUrl & name );
  QTextDocument * cache();
  
 protected:
  QTextDocument * resources;
    virtual void keyPressEvent ( QKeyEvent * e );
  QWidget * dest;
};




//! UMsgTab class contains all widgets needed by one single connection 
/*! It consists of a command line, and text fields (for incoming 
    and outgoing messages) that are all wrapped by lower level classes
    UMsgOutPane and UMsgInPane 

    It forwards messages from UMsgOutPane to URobot object, and receives
    messages from URobot object and parses them.

    Sound messages are played (if allowed by user settings), others are
    displayed in a the text field in UMsgInPane.

    Receiving a msg is done in following manner:
    Either calling recieveMsg function directly, 
    !!! it must be done from the Qt GUI-thread;
    or posting an event within QApplication, specifying instance of this class 
    as a receiver, and passing UCustomEvent as event with valid UMessage object

*/
class UMsgTab : public QWidget {
  
  Q_OBJECT
  
 public:

  UMsgTab( QWidget* parent = NULL);
  ~UMsgTab();

  virtual bool event           ( QEvent* e );

  void         send            ( QString& command );
  void         receiveMsg      ( UMessage& msg );
  
  void         setRobot        ( URobot* robot );
  URobot*      getRobot        () const;
  const char*  getRobotHostName() const;
  char*        getConnId       () ;
   
  void         clientError     ( URobot::UClientErrType type );


 public slots:
 virtual void restoreInPane();
 private:

  void saveImage   ( QString &imgName );

  void loadSettings();
  void saveSettings();

  UMsgInPane*  msgInPane;       ///< widget containing incoming messages text field
  //Q3TextEdit*  msgInTextEdit;   ///< shortcut to incoming messages text field
  UCacheTextBrowser*  msgInTextEdit;   ///< shortcut to incoming messages text field
  UMsgOutPane* msgOutPane;      ///< widget containing outgoing messages text field and command line
  QVBoxLayout* msgTabLayout;    ///< layout
  QSplitter*   msgPaneSplitter; ///< sliding splitter

  URobot       *robot;          ///< robot object that which this tab operates with

  unsigned int imgId;           ///< unique suffix of a image name displayed in a tab's text field
  unsigned int soundId;         ///< unique suffix of a sound name displayed in a tab's text field

  QStringList  tmpFiles;        ///< list of names of temporary files
  QString      tmpDirName;      ///< name of the directory where temporary files are stored
  QTime        startTime;
  QMap<int,
       QString>lineNum2ImgName; ///< list of names of images with the number of the line where they are in the text field,
                                ///< workaround for saving an image
  QString      inPaneHTML;
  int          inPanePos;
  /* user settings */
  bool playSoundOnReceive;      ///< if true sound is played when received (if NAS evailable) 

 private slots:
 virtual void anchorClicked( const QUrl &link); ///< called when an anchor in msgInTextEdit is clicked
  //virtual void textEditClicked( int para, 
  //int pos ); ///< called when msgInTextEdit is clicked 
  /////< image saving workaround
};


#endif // #ifndef UMSGTAB_H
