/*! file urobot.cpp
*******************************************************************************

File: urobot.cppn
Implementation of the URobot class.

This file is part of
%URBI Lab, version __urbilabversion__n
(c) Jean-Christophe Baillie, 2004-2005.

Permission to use, copy, modify, and redistribute this software for
non-commercial use is hereby granted.

This software is provided "as is" without warranty of any kind,
either expressed or implied, including but not limited to the
implied warranties of fitness for a particular purpose.

For more information, comments, bug reports: http://urbi.sourceforge.net

**************************************************************************** */


#include <qstring.h>
#include <qsettings.h>
#include <qdatetime.h>
   //Added by qt3to4:


#include "urobot.hh"
#include "uconn.hh"
#include "udebugtab.hh"
#include "umsgtab.hh"
#include "ugraph.hh"


const unsigned int URobot::MSG_PORT = 54000;
const unsigned int URobot::DEBUG_PORT = 59000;
const char* URobot::TAG_CONN_ID = "ident";

//! URobot constructor.
/*!
  Constructs the URobot object taking hostname as parameter

  param name is hostname of the robot

*/
URobot::URobot( const char *name )
  : QObject(  ) {

    loadSettings();

    hostName = strdup(name);

    debugClient =  0;

    // set AutoDelete ON, so the objects in the lists will be deleted automatically when list is deleted

    uid = 0;
  }

//! URobot destructor.
/*!
  Destroys all the clients and saves settings

*/
URobot::~URobot( ) {

  saveSettings();

  closeDebugClient( );
  free(hostName);

}

//! Returns a unique URBI tag.
/*!
  Tag is in format URBILAB_HHMMSSMsMsUI
*/
QString
URobot::getUniqueTag() {
  QTime time = QTime::currentTime();

  return QString("URBILAB_")
    + QString::number( time.hour() )
    + QString::number( time.minute() )
    + QString::number( time.second() )
    + QString::number( time.msec() )
    + QString::number( ++uid );
}



//! Sets up a debug client for the robot.
/*!
  param tab is pointer to the tab that demands the debug client
*/
void
URobot::setupDebugClient( UDebugTab* tab) {

  if( debugClientExists( ) ) {
    qDebug("URobot::setupDebugClient: debugClient for %s already present..returning...", hostName);
    return;
  }

  debugTab = tab;

  qDebug("URobot::setupDebugClient: creating debugClient for %s...", hostName);
  debugClient = new QTcpSocket();
  connect( debugClient, SIGNAL(connected()), SLOT(debugClientConnected()) );
  connect( debugClient, SIGNAL(error((URobot::DebugClientErrType)int)), SLOT(debugClientError(int)) );
  connect( debugClient, SIGNAL(connectionClosed()), SLOT(debugClientClosed()) );
  connect( debugClient, SIGNAL(readyRead()), SLOT(debugClientReadyRead()) );

  qDebug("URobot::setupDebugClient: connecting to host %s...", hostName);
  debugClient->connectToHost( hostName, DEBUG_PORT );
}


//! Closes robot's debug client.
void
URobot::closeDebugClient() {

  if( debugClient == 0 )
    return;
  qDebug("URobot::closeDebugClient: closing and deleting debug client for %s", hostName);
  //debugClient->clearPendingData( );
  debugClient->close( );
  delete debugClient;
  debugClient = 0;
}

//! Slot activated when debug client succesfully connected.
void
URobot::debugClientConnected() {
  qDebug("URobot::debugClientConnected: debug client connected for %s", hostName);
  debugTab->clientConnected( );
}

//! Slot activated when debug client could not connect.
void
URobot::debugClientError( QAbstractSocket::SocketError error ) {
  switch ( error ) {
  case QTcpSocket::ConnectionRefusedError:
    qWarning( "URobot::debugClientError: connection refused by %s...", hostName );
    break;
  case QTcpSocket::HostNotFoundError:
    qWarning( "URobot::debugClientError: host %s not found...", hostName );
    break;
  case QTcpSocket::RemoteHostClosedError:
    qWarning( "URobot::debugClientError: failed to read from debugSocket of %s...", hostName );
    break;
  default:
    Q_ASSERT( true );
  }

  debugTab->clientError( (DebugClientErrType) error );
}


//! Slot activated when other side closed the client.
void
URobot::debugClientClosed() {
  qWarning( "URobot::debugClientClosed: debug connection closed by server %s...", hostName );
  debugTab->clientClosed( );
}


//! Slot activated when new data arrived to the debug client.
void
URobot::debugClientReadyRead() {
  //qDebug( "URobot::debugClientReadyRead: ready to read socket: %s...", hostName );
  while( debugClient->canReadLine() ) {
    QString tmp = debugClient->readLine();
    debugTab->append( tmp );
  }
}


//! Sets up a msg client for the robot and a particular tab.
/*!
  param tab is pointer to the tab that demands the msg client
*/
void
URobot::setupMsgClient( UMsgTab* tab ) {
  qDebug("URobot::setupMsgClient: setting up a msg client for %s...", hostName);
  UClient *client = new UClient( hostName, MSG_PORT );

  if( client->error() ) {
    qWarning("URobot::setupMsgClient: msg client failed to connect to %s...", hostName);
    tab->clientError( MsgClientErr ); //propagate error to the tab
  }

  // creates the connection wrapper
  UConn* conn = new UConn( client, tab );

  // registers callback, so after the UConn receives its connection id
  client->setCallback( callback(*conn, &UConn::setConnId), TAG_CONN_ID );

  // set callback function for all received msgs
  client->setWildcardCallback( callback(*conn, &UConn::receive) );

  // put to client on the list
  msgClientList.append( conn );

  // when creating a msg client if there's no graph connection for the robot, we create one
  if( !graphClientExists( ) ) {
    qDebug("URobot::setupMsgClient: graph client for %s doesnt exist yet", hostName);
    if( setupGraphClient( ) == 0)
      qDebug("URobot::setupMsgClient: graph client for %s cannot be set up", hostName);
    //TODO:tell user
  }
}

//! Closes robot's msg client corresponding to one particular tab.
/*! We find the UConn opened for the tab and we delete it
  along with the UClient encapsuled in UConn

  param tab is the tab that wants its msg client to be closed
*/
void
URobot::closeMsgClient( UMsgTab* tab ) {
  UConn* conn=0;
  for ( QLinkedList<UConn*>::iterator it = msgClientList.begin(); 
        it != msgClientList.end() ;
        it++) {
        if ((*it)->tab == tab) {
           qDebug("URobot::closeMsgClient: deleting UConn object for %s...", hostName);
           conn = *it;
           msgClientList.erase(it);
           delete conn;
           break;
           }
        }

}


//! Returns pointer to the message tab according to UClient object that the tab is using.
/*! client is the client that the tab is using
 */
UMsgTab*
URobot::getMsgTabByMsgClient( UClient* client ) {
                              
  for ( QLinkedList<UConn*>::iterator it = msgClientList.begin(); 
        it != msgClientList.end() ;
        it++) 
        if ((*it)->client == client)
           return (*it)->tab;
  return 0;
}


//! Forward message from the tab to the robot
/*!
  param tab is pointer to the tab that is sending the message
  param msg is the message to be sent
*/
void
URobot::send( UMsgTab* tab, QString& msg ) {
  qDebug("URobot::send: sending cmd (%s) to %s...", msg.toAscii().constData(), hostName );
  getMsgClientByMsgTab( tab )->send( msg.toAscii() );
}


//! Returns UClient object used by particular tab.
/*!
  param tab is pointer to the tab that uses the client
*/
UClient*
URobot::getMsgClientByMsgTab( UMsgTab* tab ) {                                                         
  for ( QLinkedList<UConn*>::iterator it = msgClientList.begin(); 
        it != msgClientList.end() ;
        it++) 
        if ((*it)->tab == tab)
           return (*it)->client;
  return 0;
}
 


//! Returns connection id corresponding to a particular tab
/*!
  param tab is pointer to the tab whose connection id we want to know
*/
char*
URobot::getMsgConnIdByMsgTab( UMsgTab* tab ) const {                               
  for ( QLinkedList<UConn*>::const_iterator it = msgClientList.begin(); 
        it != msgClientList.end() ;
        it++) 
        if ((*it)->tab == tab)
           return (*it)->id;
  return 0;
}



//! Sets up a graph client for the robot and returns it.
UClient*
URobot::setupGraphClient( ) {
  qDebug("URobot::setupGraphClient: setting up a graph client for %s...", hostName);
  UClient *client = new UClient( hostName, MSG_PORT );

  if( client->error( ) ) {
    qWarning("URobot::setupGraphClient: graph client failed to connect to %s...", hostName);
    //TODO:tell the user
    return 0;
  }

  UConn* conn = new UConn( client, NULL, UConn::GraphConn );
  client->setCallback( callback(*conn, &UConn::setConnId), TAG_CONN_ID ); //determine the connection ID
  graphClientList.prepend( conn );
  return client;
}


//! Returns a suitable graph client for the robot.
/*!
  Suitable means a graph client that is used for graphing less that a certain
  number of variables.
*/
UClient*
URobot::getGraphClient( ) {
   UClient* client = 0;
   for ( QLinkedList<UConn*>::iterator it = graphClientList.begin(); 
        it != graphClientList.end() ;
        it++) 
         if( (*it)->graphedVariablesCnt < maxGraphClientVariables ) {
              qDebug( "URobot::getGraphClient: found old suitable connection..." );
              client = (*it)->client;
              break;
              }
         
  // if we found no suitable connection in the list, we set up a new one
  if( client == 0 )
    client = setupGraphClient();

  return client;
}


//! Increase a number of variables graphed by on UClient
/*!
  param client is UClient object whose number of graphed variables should be increased
*/
void
URobot::incGraphedVariablesCnt( UClient* client ) {
     for ( QLinkedList<UConn*>::iterator it = graphClientList.begin(); 
        it != graphClientList.end() ;
        it++) 
        if( (*it)->client == client ) {
            Q_ASSERT( (*it)->graphedVariablesCnt < maxGraphClientVariables );
            (*it)->graphedVariablesCnt++;
            break;
            }
}

//! Decrease a number of variables graphed by on UClient
/*!
  param client is UClient object whose number of graphed variables should be decreased
*/
void
URobot::decGraphedVariablesCnt( UClient* client ) {
     for ( QLinkedList<UConn*>::iterator it = graphClientList.begin(); 
        it != graphClientList.end() ;
        it++) 
        if( (*it)->client == client ) {
            Q_ASSERT( (*it)->graphedVariablesCnt >0 );
            (*it)->graphedVariablesCnt--;
            break;
            }
}


//! Function takes graphing demands.
/*! Function makes up a unique id(tag) of the graphing process
  and ask the robot for the type of the variable and
  sets a callback as a receiver of the variable type
  (who further manages the graphing process).
*/
void
URobot::startGraph( QString &variable ) {
  qDebug( "URobot::startGraph: demand for graphing variable: %s", variable.toAscii().constData() );

  // client that is used just for asking for the variable type
  UClient* client = getGraphClient( );

  if( client == 0 )
    qDebug( "URobot::startGraph: graphing not possible, cannot set up graphing client" );
  //TODO: tell user

  // unique tag that is used just for asking for the variable type
  QString uniqueTag = getUniqueTag();


  qDebug( "URobot::startGraph: asking for type, using tag: %s", uniqueTag.toAscii().constData() );
  // callback used for determining the variable type
  // we have to force the template creation
  client->setCallback( callback<URobot,QString&>( *this, &URobot::setGraphCallback, variable ), uniqueTag.toAscii() );

  //qDebug( "URobot::startGraph: asking for type, sending to robot: " + uniqueTag + ": " + variable );
  // we request a variable value (along with it we receive the type)
  client->send( uniqueTag.append(": " + variable + ";").toAscii() );
}


//! Function sets the callback used for graphing.
/*! receives a variable value, according to variable type chooses type of client(graph, multimedia)
  and ask robot to send the value of the variable in a cycle;
  registers the callback UGraph::plot for it
  puts the (uniqueTag, callbackId) in the list, so the callback can be unregistered in the future.

  param variable is name of the graphed variable
  param msg is the message containing the value of the variable
*/
UCallbackAction
URobot::setGraphCallback( QString& variable, const UMessage& msg ) {
  UClient* client = 0;
  UCallbackID callbackId = 0;
  int period = 1000;                     // update period (1/frequency) [millisec]
  QString uniqueTag = getUniqueTag();    // unique tag used for graphing of the variable
  bool typeValid = false;                // is the type of variable valid for graphing?


  switch(msg.type) {
    /***** double ****/
  case MESSAGE_DOUBLE:
    //qDebug("URobot::setGraphCallback: request for graphing variable: " + variable + "(DOUBLE) received..." );
    typeValid = true;
    client = getGraphClient( );

    if( client == 0 ) {
      qDebug( "URobot::startGraph: graphing not possible, cannot set up graphing client" );
      //TODO: tell user
      break;
    }

    // set the callback for msgs with variable values
    incGraphedVariablesCnt( client );
    callbackId = client->setCallback( callback(graph, &UGraph::plotEventDouble ), uniqueTag.toAscii() );
    period = 30;
    break;

    /***** string ****/
  case MESSAGE_STRING:
    qDebug("URobot::setGraphCallback: attemping to graph a STRING variable...failed...");
    //TODO: tell user
    break;

    /***** binary ****/
  case MESSAGE_BINARY:

    switch(msg.binaryType) {
      /***** bin none ****/
    case BINARYMESSAGE_NONE:
      //will never happen
      qDebug("URobot::setGraphCallback: attemping to graph a BINARY(NONE) variable...failed...");
      //TODO: tell user
      break;

      /***** bin unknown ****/
    case BINARYMESSAGE_UNKNOWN:
      qDebug("URobot::setGraphCallback: attemping to graph a BINARY(UNKNOWN) variable...failed...");
      //TODO: tell user
      break;

      /***** bin image ****/
    case BINARYMESSAGE_IMAGE:
      //qDebug("URobot::setGraphCallback: request for graphing variable: " + variable + "(IMAGE) received...");
      typeValid = true;
      client = getAVClient( );
      if( client == 0 ) {
	qDebug( "URobot::startGraph: graphing not possible, cannot set up graphing client" );
	//TODO: tell user
	break;
      }
      // set the callback for msgs with variable values
      callbackId = client->setCallback( callback(graph, &UGraph::plotEventImage ), uniqueTag.toAscii() );
      period = 30;
      break;

      /***** bin sound ****/
    case BINARYMESSAGE_SOUND:
      //qDebug("URobot::setGraphCallback: request for graphing variable: " + variable + "(SOUND) received...");
      typeValid = true;
      client = getAVClient( );
      if( client == 0 ) {
	qDebug( "URobot::startGraph: graphing not possible, cannot set up graphing client" );
	//TODO: tell user
	break;
      }

      // set the callback for msgs with variable values
      callbackId = client->setCallback( callback(graph, &UGraph::plotEventSound ), uniqueTag.toAscii() );
      period = 1000;
      break;

    default:
      Q_ASSERT(true);
    } // switch(msg.binaryType) {

    break;

    /***** system ****/
  case MESSAGE_SYSTEM:
    qDebug("URobot::setGraphCallback: attemping to graph a variable...but SYSTEM msg received...");
    //TODO: tell user
    break;

    /***** error ****/
  case MESSAGE_ERROR:
    qDebug("URobot::setGraphCallback: attemping to graph a variable, that is not declared...");
    //TODO: tell user variable not declared
    break;

    /***** unknown ****/
  case MESSAGE_UNKNOWN:
    qDebug("URobot::setGraphCallback: attemping to graph a variable...but UNKNOWN msg received...");
    //TODO: tell user
    break;

  default:
    Q_ASSERT(true);

  } // switch(msg.type){

  if( typeValid ) {

    // stores (uniqueTag,callbackId) in a list
    callbackIdWrapperList.insert( uniqueTag, UCallbackIDWrapper( callbackId, client ) );

    // registers the robot, uniqueTag, variable name and type in the UGraph object as a graph process
    graph.registerGraphProcess( this, uniqueTag, variable, msg.type, msg.binaryType );

    // we send a command to the robot that tells him to send us the value every X microseconds
    // uniquetag: loop{ uniquetag: var; wait(delay) },
    QString command = uniqueTag
      + ": loop{" + uniqueTag + ":" + variable
      + ";wait(" + QString::number( period ) + ")},";

    //qDebug( "URobot::setGraphCallback: starting the values stream, sending to robot: " + command );
    client->send( command.toAscii() );
  }
  return URBI_REMOVE; // we unregister the callback( that is used just once to determine the type of the variable)

}

//! Stop the graphing, using the graphing process' tag.
void
URobot::stopGraph( QString& uniqueTag ) {
  //qDebug("URobot::stopGraph: stopping graphing assoc. with tag " + uniqueTag );
  // we look up the client and callbackId associated with the tag
  UCallbackIDWrapper wrap = callbackIdWrapperList.value( uniqueTag );

  if( wrap.client == 0 ) {
    qWarning( "URobot::stopGraph: trying to stop graphing that is not registered with the robot..." );
    return;
  }

  // we send the robot a command that stops sending of the values
  QString stopCmd = "stop ";
  stopCmd += uniqueTag + ";";
  qDebug("URobot::stopGraph: sending to the robot: %s" ,stopCmd.toAscii().constData() );
  getGraphClient()->send( stopCmd.toAscii() ); // wrap->client might be flooded so we use a different free client
  // we unregister the callback
  wrap.client->deleteCallback( wrap.callbackId );

  // finally we remove the wrapper from the list
  callbackIdWrapperList.take( uniqueTag );

  decGraphedVariablesCnt( wrap.client );


}

//! Returns a multimedia client for the robot.
UClient*
URobot::getAVClient( ) {
  qDebug("URobot::setupAVClient: setting up a multimedia client for %s...", hostName);
  UClient *client = new UClient( hostName, MSG_PORT );

  if( client->error( ) ) {
    qWarning("URobot::setupAVClient: multimedia client failed to connect to %s...", hostName);
    delete client;
    return 0;
    //TODO:tell the user, retry???
  }

  UConn* conn = new UConn( client, 0, UConn::AVConn );
  client->setCallback( *conn, &UConn::setConnId, TAG_CONN_ID ); //determine the connection ID
  AVClientList.append( conn );

  return client;
}


//! Loads user settings.
void
URobot::loadSettings() {
  maxGraphClientVariables = settings.value( APP_KEY + "/robot/maxGraphClientVariables", 10 ).toInt();

}


//! Saves user settings.
void
URobot::saveSettings() {
  settings.setValue( APP_KEY + "/robot/maxGraphClientVariables", maxGraphClientVariables );

}






