/*! \file urobot.hh
 *******************************************************************************

 File: urobot.hh\n
 Definition of the URobot class.

 This file is part of 
 %URBI Lab, version __urbilabversion__\n
 (c) Jean-Christophe Baillie, 2004-2005.

 Permission to use, copy, modify, and redistribute this software for
 non-commercial use is hereby granted.

 This software is provided "as is" without warranty of any kind,
 either expressed or implied, including but not limited to the
 implied warranties of fitness for a particular purpose.

 For more information, comments, bug reports: http://urbi.sourceforge.net

 **************************************************************************** */

#ifndef UROBOT_H
#define UROBOT_H

#include <qlist.h>
#include <qobject.h>
#include <qhash.h>
#include <QLinkedList>
#include <QTcpSocket>

#include <uclient.h>  // liburbi

#include "uconn.hh"
 
class QString;
class QSettings;

class UDebugTab;
class UMsgTab;
class UConn;
class UGraph;


extern const QString   APP_KEY;  ///< application key
extern       QSettings settings; ///< contains the user settings
extern       UGraph    graph;    ///< graphing device


//! URobot class represent one robot.
/*! It provides the ways to communicate with robot.
    It contains message, debug, graphing, sound and video connection.
    All connection except for debug (using QSocket) are done by means of liburbi.


*/
class URobot : public QObject {

  Q_OBJECT
 
 public:

  enum UClientErrType {
    MsgClientErr=0,
    GraphClientErr
  }; ///< type message or graph error 

  enum DebugClientErrType {
    ConnRefusedErr = QTcpSocket::ConnectionRefusedError,
    HostNotFoundErr = QTcpSocket::HostNotFoundError,
    SocketReadErr = QTcpSocket::RemoteHostClosedError
  }; ///< type of debug connection error


  URobot( const char *name );
  ~URobot( );
  
  const char* getHostName () const { return hostName; };

  QString     getUniqueTag();
 

  // debug connection  
  void       setupDebugClient ( UDebugTab* tab);
  bool       debugClientExists( ) const { return debugClient != 0; };
  UDebugTab* getDebugTab      ( ) const { return debugTab; } ;
  void       closeDebugClient ( );               


  // message connections 
  void     setupMsgClient      ( UMsgTab* tab );                 
  void     closeMsgClient      ( UMsgTab* tab );                 
  UMsgTab* getMsgTabByMsgClient( UClient* client );    
  UClient* getMsgClientByMsgTab( UMsgTab* tab );
  char*    getMsgConnIdByMsgTab( UMsgTab* tab ) const;
  void     send                ( UMsgTab* tab, QString &msg );             

  
  // graph connections 
  UClient* setupGraphClient ();                  
  bool     graphClientExists() { return !graphClientList.empty(); };
  UClient* getGraphClient   ();
  void     startGraph       ( QString& var );   
  void     stopGraph        ( QString& uniqueTag );
  UCallbackAction setGraphCallback( QString& variable, 
				    const UMessage& msg ); 
 
  // multimedia connections
  UClient* getAVClient();                     


 private:

  ///> for internal use, wraps UCallbackID and client that the callback is registered at
  class UCallbackIDWrapper {
  public:
     UCallbackIDWrapper() : callbackId(0), client(0) {}
    UCallbackIDWrapper( UCallbackID id, UClient* cl ) 
      : callbackId(id), client(cl){ };
    UCallbackID callbackId;
    UClient* client;
  };

  void incGraphedVariablesCnt( UClient* client );
  void decGraphedVariablesCnt( UClient* client );

  void loadSettings(); 
  void saveSettings(); 

  
  char* hostName;                  ///< server name

  int uid;                               ///< suffix for unique tag
  
  QTcpSocket*   debugClient;                ///< connection for receiving debug messages
  UDebugTab* debugTab;                   ///< tab corresponding to the debugClient

  QLinkedList<UConn*> msgClientList;        ///< connections for message interchange
  QLinkedList<UConn*> graphClientList;      ///< connections for variable tracking purposes (graphs, etc.)
  QLinkedList<UConn*> AVClientList;         ///< multimedia connections (video, audio) 
  
  QHash<QString, UCallbackIDWrapper> callbackIdWrapperList; ///< list of callback ids and associated clients, addressed by unique tag
  
  // user settings 
  int maxGraphClientVariables; ///< max.number of variables simultaneously graphed by one graphClient

 
  static const unsigned int MSG_PORT;       ///< port at which the robot listens for commands and send replies
  static const unsigned int DEBUG_PORT;     ///< port at which the robot sends debug info
  static const char* TAG_CONN_ID;           ///< name of the tag which comes the connection id with

   
 private slots:
 
  void debugClientConnected( );        ///< called by socket debugClient when connection is ready
  void debugClientError    ( QAbstractSocket::SocketError error );  ///< called by socket debugClient when there's an connection error
  void debugClientClosed   ( );        ///< called by socket debugClient when the connection is closed 
  void debugClientReadyRead( );        ///< called by socket debugClient when there's data in the input buffer
  
};


#endif // #ifndef UROBOT_H
