/*
 * Copyright (C) Gostai S.A.S., 2007.  All rights reserved.
 *
 * This software is provided "as is" without warranty of any kind,
 * either expressed or implied, including but not limited to the
 * implied warranties of fitness for a particular purpose.
 *
 * See the LICENSE file for more information.
 * For comments, bug reports and feedback: http://www.urbiforge.com
 */
#include "houghlines.hh"

/**
 * \brief if hack is set to true, only the lines
 * longer than HACK_HEIGHT pixels are returned.
 */
HoughLines::HoughLines(bool hack) : hack (hack) {
  result = 0;
  storage = cvCreateMemStorage(0);
}

HoughLines::~HoughLines() {
  cvReleaseMemStorage(&storage);
}

CvSeq&
HoughLines::getResult() {
  return *result;
}

void
HoughLines::freeItems() {
  result = 0;
  cvClearMemStorage(storage);
}

void
HoughLines::compute(const IplImage& img) {
  IplImage*	tmp;

  debug && std::cout <<
    "HoughLines is computing image" << std::endl;
  freeItems();
  tmp = cvCreateImage(cvGetSize(&img), 8, 1);
  cvCvtColor(&img, tmp, CV_BGR2GRAY);
  cvCanny(tmp, tmp, 50, 200, 3);
  result = cvHoughLines2(tmp, storage, CV_HOUGH_PROBABILISTIC,
			1, CV_PI/180, 50, 50, 10);
  if (hack)
    workOnLines();
  cvReleaseImage(&tmp);
}

/**
 * \brief \arg hack: if true, only the lines
 * longer than HACK_HEIGHT pixels are returned.
 */
void
HoughLines::setHack(bool hack) {
  this->hack = hack;
}

//! called if hack is set to true
void
HoughLines::workOnLines() {
  for (int i = 0; i < result->total; ++i) {
    CvPoint* line = (CvPoint*)cvGetSeqElem(result,i);
    if (line[0].y <= HACK_HEIGHT || line[1].y <= HACK_HEIGHT ||
	pow(line[1].x - line[0].x, 2) +
	pow(line[0].y - line[1].y, 2) <= SQUARE_DST)
      cvSeqRemove(result, i--);
  }
}

void
HoughLines::writeOn(IplImage& img) const {
  debug && std::cout <<
    "HoughLines is writing on image" << std::endl;
  if (!result)
    std::cerr << "HoughLines::writeOn(img) called before HoughLines::compute(img)" << std::endl;
  for (int i = 0; i < result->total; ++i) {
    CvPoint* line = (CvPoint*)cvGetSeqElem(result,i);
    cvLine(&img, line[0], line[1], CV_RGB(255, 100, 100), 2, 8);
  }
}

void
HoughLines::emptyResult() {
  freeItems();
}
