/*
 * Copyright (C) Gostai S.A.S., 2007.  All rights reserved.
 *
 * This software is provided "as is" without warranty of any kind,
 * either expressed or implied, including but not limited to the
 * implied warranties of fitness for a particular purpose.
 *
 * See the LICENSE file for more information.
 * For comments, bug reports and feedback: http://www.urbiforge.com
 */
#include "squares.hh"

Squares::Squares() {
  result = 0;
  thresh = 50;
  storage = cvCreateMemStorage(0);
}

Squares::~Squares() {
  if (storage)
    cvReleaseMemStorage(&storage);
  storage = 0;
}

void
Squares::emptyResult() {
  cvClearMemStorage(storage);
  result = 0;
}

/**
 * \brief helper function:
 * finds a cosine of angle between vectors
 * from pt0->pt1 and from pt0->pt2 
 */
double
Squares::angle(CvPoint* pt1, CvPoint* pt2, CvPoint* pt0)
{
    double dx1 = pt1->x - pt0->x;
    double dy1 = pt1->y - pt0->y;
    double dx2 = pt2->x - pt0->x;
    double dy2 = pt2->y - pt0->y;
    return (dx1*dx2 + dy1*dy2)/sqrt((dx1*dx1 + dy1*dy1)*(dx2*dx2 + dy2*dy2) + 1e-10);
}

void
Squares::compute(const IplImage& img) {
  CvSeq* contours;
  int i, c, l, N = 11;
  IplImage* tmp = cvCloneImage(&img);
  CvSize sz = cvSize(img.width & -2, img.height & -2);
  IplImage* gray = cvCreateImage(sz, 8, 1); 
  IplImage* pyr = cvCreateImage(cvSize(sz.width/2, sz.height/2), 8, 3);
  IplImage* tgray;
  double s, t;

  emptyResult();
  result = cvCreateSeq(0, sizeof(CvSeq), sizeof(CvPoint), storage);
  // select the maximum ROI in the image
  // with the width and height divisible by 2
  cvSetImageROI(tmp, cvRect(0, 0, sz.width, sz.height));
  // down-scale and upscale the image to filter out the noise
  cvPyrDown(tmp, pyr, 7);
  cvPyrUp(pyr, tmp, 7);
  tgray = cvCreateImage(sz, 8, 1);
  // find squares in every color plane of the image
  for (c = 0; c < 3; c++) {
    // extract the c-th color plane
    cvSetImageCOI(tmp, c+1);
    cvCopy(tmp, tgray, 0);
    // try several threshold levels
    for (l = 0; l < N; l++) {
      // hack: use Canny instead of zero threshold level.
      // Canny helps to catch squares with gradient shading   
      if (l == 0) {
	// apply Canny. Take the upper threshold from slider
	// and set the lower to 0 (which forces edges merging) 
	cvCanny(tgray, gray, 0, thresh, 5);
	// dilate canny output to remove potential
	// holes between edge segments 
	cvDilate(gray, gray, 0, 1);
      }
      else {
	// apply threshold if l!=0:
	//     tgray(x,y) = gray(x,y) < (l+1)*255/N ? 255 : 0
	cvThreshold(tgray, gray, (l+1)*255/N, 255, CV_THRESH_BINARY);
      }
      // find contours and store them all as a list
      cvFindContours(gray, storage, &contours, sizeof(CvContour),
		     CV_RETR_LIST, CV_CHAIN_APPROX_SIMPLE,
		     cvPoint(0,0));
      // test each contour
      while(contours) {
	// approximate contour with accuracy proportional
	// to the contour perimeter
	result = cvApproxPoly(contours, sizeof(CvContour), storage,
			      CV_POLY_APPROX_DP, cvContourPerimeter(contours)*0.02, 0);
	// square contours should have 4 vertices after approximation
	// relatively large area (to filter out noisy contours)
	// and be convex.
	// Note: absolute value of an area is used because
	// area may be positive or negative - in accordance with the
	// contour orientation
	if (result->total == 4 &&
	    fabs(cvContourArea(result,CV_WHOLE_SEQ)) > 1000 &&
	    cvCheckContourConvexity(result)) {
	  s = 0;
	  for (i = 0; i < 5; i++) {
	    // find minimum angle between joint
	    // edges (maximum of cosine)
	    if (i >= 2) {
	      t = fabs(angle(
			     (CvPoint*)cvGetSeqElem(result, i),
			     (CvPoint*)cvGetSeqElem(result, i-2),
			     (CvPoint*)cvGetSeqElem(result, i-1)));
	      s = s > t ? s : t;
	    }
	  }
	  // if cosines of all angles are small
	  // (all angles are ~90 degree) then write quandrange
	  // vertices to resultant sequence 
	  if (s < 0.3)
	    for (i = 0; i < 4; i++)
	      cvSeqPush(result,
			(CvPoint*)cvGetSeqElem(result, i));
	}
	// take the next contour
	contours = contours->h_next;
      }
    }
  }
  // release all the temporary images
  cvReleaseImage(&gray);
  cvReleaseImage(&pyr);
  cvReleaseImage(&tgray);
  cvReleaseImage(&tmp);
}

void
Squares::actOnKeyPressed(char& key) {
}

void
Squares::writeOn(IplImage& img) const {
    CvSeqReader reader;
    int i;
    
    // initialize reader of the sequence
    cvStartReadSeq(result, &reader, 0);
    
    // read 4 sequence elements at a time (all vertices of a square)
    for (i = 0; i < result->total; i += 4)
    {
        CvPoint pt[4], *rect = pt;
        int count = 4;
        
        // read 4 vertices
        CV_READ_SEQ_ELEM(pt[0], reader);
        CV_READ_SEQ_ELEM(pt[1], reader);
        CV_READ_SEQ_ELEM(pt[2], reader);
        CV_READ_SEQ_ELEM(pt[3], reader);
        
        // draw the square as a closed polyline 
        cvPolyLine(&img, &rect, &count, 1, 1, CV_RGB(0,255,0), 3, CV_AA, 0);
    }
}


CvSeq&
Squares::getResult() {
  return *result;
}
