/*
 * Copyright (C) Gostai S.A.S., 2007.  All rights reserved.
 *
 * This software is provided "as is" without warranty of any kind,
 * either expressed or implied, including but not limited to the
 * implied warranties of fitness for a particular purpose.
 *
 * See the LICENSE file for more information.
 * For comments, bug reports and feedback: http://www.urbiforge.com
 */
#include "vision.hpp"

//public

std::string	Vision::window_name = "window_name";

/**
 * \arg delay: Use it to skip frames.
 * 
 * By example: delay = 2. You call workOn(img1) once.
 * Then you called it again workOn(img2)
 * and you get img1 (no computations have been done)
 * and then you called it with workOn(img3) and here
 * you get img3...
 *
 * \arg night: If true, you see only the result
 * of the computations in the window
 */
Vision::Vision(int delay, bool night) :
  delay (delay), night (night) {
  result = 0;
  wait = 0;
}

/**
 * \arg behaviors: the list of VisionBehavior classes
 * that the Vision driver will manage
 * 
 * \arg delay: Use it to skip frames.
 * 
 * By example: delay = 2. You call workOn(img1) once.
 * Then you called it again workOn(img2)
 * and you get img1 (no computations have been done)
 * and then you called it with workOn(img3) and here
 * you get img3...
 *
 * \arg night: If true, you see only the result
 * of the computations in the window
 */
Vision::Vision(VisionBehaviors& behaviors, int delay, bool night) :
  delay (delay), behaviors (behaviors), night (night) {
  result = 0;
  wait = 0;
}

Vision::~Vision() {
  freeResult();
}

/**
 * \arg delay: Use it to skip frames.
 * 
 * By example: delay = 2. You call workOn(img1) once.
 * Then you called it again workOn(img2)
 * and you get img1 (no computations have been done)
 * and then you called it with workOn(img3) and here
 * you get img3...
 */
void
Vision::setDelay(int delay) {
  this->delay = delay;
}

/**
 * \arg night: If true, you see only the result
 * of the computations in the window
 */
void
Vision::setNight(bool b) {
  night = b;
}

/**
 * \brief Use an OpenCv CvCapture to get frames,
 * do the computations on them, and then
 * show the results on a window.
 */
void
Vision::workAndDisplay(CvCapture& camera) {
  IplImage*	frame = 0;
  IplImage	img;

  cvNamedWindow(window_name.c_str(), 1);
  while (cvGetWindowHandle(window_name.c_str())) {
    frame = cvQueryFrame(&camera);
    img = workOn(*frame);
    cvShowImage(window_name.c_str(), &img);
    keyPressed = cvWaitKey(2);
    actOnKeyPressed();
  }
}

/**
 * \brief Use the Vision driver on a single Image.
 * 
 * \return an image, don't free it!
 */
IplImage&
Vision::workOn(const IplImage& src) {
  freeResult();
  result = cvCloneImage(&src);
  return workOn2(*result);
}

//private

IplImage&
Vision::workOn2(IplImage& result) {
  VariantImgArg source(result);

  if (!delay || !(wait++ % delay))
    BOOST_FOREACH(Variant i, behaviors)
      boost::apply_visitor(VBVisitor<COMPUTE>(), i, source);
  if (night)
    cvZero(&result);
  BOOST_FOREACH(const Variant i, behaviors)
    boost::apply_visitor(VBVisitor<WRITEON>(), i, source);
  return result;
}

void
Vision::actOnKeyPressed() {
  VariantKeyArg	myKey(keyPressed);

  switch (keyPressed) {
  case 27:
    cvDestroyWindow(window_name.c_str());
    break;
  case 'n':
    night = !night;
    break;
  default:
    ;
  }
  BOOST_FOREACH(Variant i, behaviors)
    boost::apply_visitor(VBKeyVisitor<KEYPRESSED>(), i, myKey);
}

void
Vision::die(std::string str) {
  std::cerr << "Error: " << str << std::endl;
  exit(1);
}

void
Vision::freeResult() {
  if (result)
    cvReleaseImage(&result);
}
