/*
 * Copyright (C) Gostai S.A.S., 2007.  All rights reserved.
 *
 * This software is provided "as is" without warranty of any kind,
 * either expressed or implied, including but not limited to the
 * implied warranties of fitness for a particular purpose.
 *
 * See the LICENSE file for more information.
 * For comments, bug reports and feedback: http://www.urbiforge.com
 */
#ifndef VISION_HH_
#define VISION_HH_

#include <iostream>
#include <cv.h>
#include <cxcore.h>
#include <highgui.h>
#include <math.h>
#include <vector>
#include <string>
#include <boost/foreach.hpp>
#include <boost/variant.hpp>
#include "visionbehavior.hpp"
#include "vbvisitor.hpp"
#include "vbkeyvisitor.hpp"


/*! \mainpage Vision Documentation
 *
 * \section intro_sec Introduction
 *
 * I am happy to see your interest for my program.
 * Don't forget to install OpenCv and Boost libraries.
 *
 * Now you just have to add your own VisionBehavior classes
 * which would be managed easily with the Vision driver class,
 * thanks to the OpenCv structures (CvCapture and IplImage).
 * 
 * You can still mail me if you want more informations:
 * cedric dot puig at gmail dot com.
 *
 * \section license_seq GPL License
 *
 * Copyright (C) Gostai S.A.S., 2007.  All rights reserved.
 *
 * This software is provided "as is" without warranty of any kind,
 * either expressed or implied, including but not limited to the
 * implied warranties of fitness for a particular purpose.
 *
 * See the LICENSE file for more information.
 * For comments, bug reports and feedback: http://www.urbiforge.com
 *
 * 
 */

typedef std::vector<CvPoint> CvPoints;

/** \brief This variant must be extended if any VisionBehavior<T> with a different T is added.
 *
 * This changement can be avoided if using a boost::recursive_variant<VisionBehavior<T>, recursive_variant<U> >.
 */
typedef boost::variant<VisionBehavior<CvSeq>*, VisionBehavior<CvPoints>*> Variant;

//! This is the type of the list of the different VisionBehaviors which with the Vision class works.
typedef std::vector<Variant> VisionBehaviors;

/** \brief The Vision class is the driver of the VisionBehavior classes.
 *
 * It can work with a single image or with a OpenCv capture device (CvCapture):
 * It means it works with a camera.
 * It can show the result of the VisionBehavior instances on a window (it is still using OpenCv).
 */
class		Vision {
private:
  IplImage*		result;
  int			delay;
  int			wait;
  char			keyPressed;
  VisionBehaviors	behaviors;
  bool			night;
  static std::string	window_name;
public:

  Vision(int delay = 0,
	 bool night = false);
  Vision(VisionBehaviors& behaviors,
	 int delay = 0,
	 bool night = false);
  ~Vision();
  template <class T>
  void			addBehavior(T* v);
  IplImage&		workOn(const IplImage&);
  void			workAndDisplay(CvCapture&);
  void			setDelay(int);
  void			setNight(bool);
private:
  IplImage&		workOn2(IplImage&);
  void			actOnKeyPressed();
  void			die(std::string);
  void			freeResult();
};


/** \brief Add a VisionBehavior on the fly
 *
 * I would prefer use addBehavior(VisionBehavior<T>*) but
 * it causes problem when I called it.
 * Nevertheless, it won't ever compile if it is not
 * this kind of type.
 */
template <class T>
void
Vision::addBehavior(T* v) {
  behaviors.push_back(v);
}

#endif /* !VISION_HH_ */
